import aioredis
import orjson
import pytest
from aioredis import RedisError
from aioredlock import LockError
from httpx import HTTPError, Request
from pytest_httpx import HTTPXMock, to_response

from peach_collector import PeachCollector, PeachCollectorQueue


@pytest.fixture()
async def redis():
    redis = await aioredis.create_redis_pool("redis://localhost")
    await redis.flushall()
    yield redis
    redis.close()
    await redis.wait_closed()


@pytest.fixture()
async def queue(redis):
    collector = PeachCollector("a-site-key", "an-app-id")
    q = PeachCollectorQueue("redis://localhost", collector, name="my_q", batch_size=2)
    for i in range(1, 6):
        await q.push_event({"foo": i})
    yield q
    await q.close_redis_conn()


@pytest.mark.asyncio
async def test_peach_collector_queue_normal(httpx_mock: HTTPXMock, redis, queue):
    # Events are sent in batches
    call_counter = 0

    def response(request, *args, **kwargs):
        nonlocal call_counter
        batches = [[1, 2], [3, 4], [5]]
        sent = orjson.loads(request.read())
        assert [e["foo"] for e in sent["events"]] == batches[call_counter]
        call_counter += 1
        return to_response(json={"ok": 1})

    httpx_mock.add_callback(response)

    await queue.drain()
    assert await redis.lrange("my_key", 0, -1) == []


@pytest.mark.asyncio
async def test_peach_collector_queue_peach_failures(httpx_mock: HTTPXMock, redis, queue):
    # First batch get through to Peach but second blows up, so those and remaining events
    # remain in queue
    call_counter = 0

    def response(request: Request, *args, **kwargs):
        nonlocal call_counter
        call_counter += 1
        if call_counter == 2:
            raise HTTPError(message="boom", request=request)
        return to_response(json={"ok": 1})

    httpx_mock.add_callback(response)

    await queue.drain()
    remaining = [orjson.loads(e) for e in await redis.lrange("my_q", 0, -1)]
    assert [e["foo"] for e in remaining] == [3, 4, 5]


@pytest.mark.asyncio
async def test_peach_collector_queue_redis_lock(queue):
    lock = await queue._acquire_lock()
    another_q = PeachCollectorQueue("redis://localhost", queue.collector, name="my_q")
    with pytest.raises(LockError):
        await another_q._acquire_lock()
    await another_q.close_redis_conn()
    await queue._release_lock(lock)

    # Now that lock should be available
    await another_q._acquire_lock()


@pytest.mark.asyncio
async def test_peach_collector_queue_redis_retries(httpx_mock: HTTPXMock, redis, queue):
    # The first redis delete call fails but the retry succeeds
    def response(request: Request, *args, **kwargs):
        return to_response(json={"ok": 1})

    httpx_mock.add_callback(response)

    call_counter = 0

    async def mock(*args, **kwargs):
        nonlocal call_counter
        call_counter += 1
        if call_counter == 1:
            raise RedisError("boom")
        return await original_ltrim(*args, **kwargs)

    original_ltrim = queue.conn.ltrim
    queue.conn.ltrim = mock

    await queue.drain()
    assert call_counter == 2

    remaining = [orjson.loads(e) for e in await redis.lrange("my_q", 0, -1)]
    assert [e["foo"] for e in remaining] == []



import logging
from typing import Dict, List, Any

import httpx

from peach_collector import __version__
from peach_collector._utils import timestamp

logger = logging.getLogger(__name__)


class PeachCollector:
    def __init__(
        self,
        site_key: str,
        app_id: str,
        server_url: str = "https://pipe-collect.ebu.io/v3/collect",
        enabled: bool = True,
        debug: bool = True,
    ):
        self.site_key = site_key
        self.app_id = app_id
        self.server_url = server_url
        self.enabled = enabled
        self.debug = debug
        self.events: List[Dict[str, Any]] = []

    def add_event(self, event):
        event = {**{"timestamp": timestamp()}, **event}
        self.events.append(event)

    async def send_events(self):
        if not self.events:
            return

        assert self.server_url and self.site_key

        payload = {
            "peach_schema_version": "1.0.3",
            "peach_implementation_version": __version__,
            "sent_timestamp": timestamp(),
            "client": {"type": "web", "app_id": self.app_id},
            "events": self.events,
        }

        if self.debug:
            logger.debug(payload)

        if self.enabled:
            logger.info(f"Sending {len(self.events)} events to peach. Url: {self.server_url}. App id: {self.app_id}")
            async with httpx.AsyncClient() as client:
                res = await client.post(self.server_url, params={"s": self.site_key}, json=payload)
                res.raise_for_status()
        else:
            logger.info(
                f"Peach collector is disabled. Would have sent {len(self.events)} events to peach"
            )

        self.events = []

    def __call__(self):
        self.send_events()



import orjson

import pytest
from pytest_httpx import HTTPXMock

from peach_collector import PeachCollector


@pytest.mark.asyncio
async def test_peach_collector(httpx_mock: HTTPXMock):
    httpx_mock.add_response(method="POST")

    collector = PeachCollector("a-site-key", "an-app-id")
    collector.add_event({"foo": 666})
    collector.add_event({"bar": 777})
    await collector.send_events()

    request = httpx_mock.get_request()
    sent = orjson.loads(request.read())
    assert len(sent["events"]) == 2
    assert sent["events"][0]["foo"] == 666
    assert sent["events"][0]["timestamp"]



import pytest
from pytest_httpx import HTTPXMock

from peach_collector._utils import retry_with_backoff, RetryException, retry_secs


class CustomException(Exception):
    pass


def test_retry_secs():
    assert retry_secs(5, 3) == [1, 8, 27, 64, 125]


@pytest.mark.asyncio
async def test_retry_with_backoff(httpx_mock: HTTPXMock):

    call_counter = 0

    def test_coro(blow_up_count):
        async def blow_up():
            nonlocal call_counter
            call_counter += 1
            if call_counter <= blow_up_count:
                raise CustomException(f"boom {call_counter}")

        return blow_up

    call_counter = 0
    await retry_with_backoff(test_coro(blow_up_count=2), 3, 1.0, CustomException)
    assert call_counter == 3

    call_counter = 0
    with pytest.raises(RetryException):
        await retry_with_backoff(test_coro(blow_up_count=2), 2, 1.0)
    assert call_counter == 2

    class AnotherException(Exception):
        pass

    call_counter = 0
    with pytest.raises(CustomException):
        # If we get an another exception than we said we stop right away
        await retry_with_backoff(test_coro(blow_up_count=3), 1, 1.0, AnotherException)
    assert call_counter == 1



import logging

import aioredis
import orjson
from aioredis import RedisError
from aioredlock import Aioredlock, Lock
from httpx import HTTPError

from ._collector import PeachCollector
from ._utils import timestamp, retry_with_backoff, RetryException, retry_secs

logger = logging.getLogger(__name__)


class PeachCollectorQueue:
    def __init__(
        self,
        redis_url: str,
        collector: PeachCollector = None,
        batch_size=200,
        name="peach_collector",
        lock_ttl_secs=60 * 60 * 24,
        delete_from_redis_retries=10,
        delete_from_redis_retries_exponent=3.0,
    ):
        self.collector = collector
        self.batch_size = batch_size
        self.redis_url = redis_url
        self.conn = None
        self.lock_manager = None
        self.name = name
        self.lock_name = f"{self.name}_lock"
        self.lock_ttl_secs = lock_ttl_secs
        self.delete_from_redis_retries = delete_from_redis_retries
        self.delete_from_redis_retries_exponent = delete_from_redis_retries_exponent
        assert (
            sum(retry_secs(delete_from_redis_retries, delete_from_redis_retries_exponent))
            < self.lock_ttl_secs
        )

    async def init_redis_conn(self):
        if not self.conn:
            self.conn = await aioredis.create_redis_pool(self.redis_url)

    async def close_redis_conn(self):
        if self.conn:
            self.conn.close()
            await self.conn.wait_closed()
            self.conn = None

    async def push_event(self, event):
        """
        Adds event to queue. Make sure it is json serializable
        """
        await self.init_redis_conn()
        event = {**{"timestamp": timestamp()}, **event}
        await self.conn.rpush(self.name, orjson.dumps(event))

    async def drain(self) -> bool:
        """
        Drains the queue and returns True if we were completely able to send all events to Peach.
        If we're only to send some events so we'll return False. Those events will then be then
        available to send on the next call to drain().

        If we're not able to delete sent events from redis we have a problem that will need to be
        manually resolved before continuing so we don't send duplicates to Peach. This will disable
        the draining from future runs until events and lock are removed manually.

        :returns A boolean indicating if we were able to send all events to Peach
        """
        logger.info("Draining")

        assert self.collector
        lock = await self._acquire_lock()

        events = await self._load_events_from_redis()
        num_sent = await self._send_to_peach(events)
        if num_sent:
            await self._delete_from_redis(num_sent)
        self._log_results(events, num_sent)

        await self._release_lock(lock)

        return num_sent == len(events)

    async def _acquire_lock(self) -> Lock:
        await self.init_redis_conn()
        if not self.lock_manager:
            urls = self.redis_url if isinstance(self.redis_url, list) else [self.redis_url]
            self.lock_manager = Aioredlock(urls)
        return await self.lock_manager.lock(self.lock_name, lock_timeout=self.lock_ttl_secs)

    async def _release_lock(self, lock: Lock):
        if self.lock_manager and lock:
            await self.lock_manager.unlock(lock)
            await self.lock_manager.destroy()
            self.lock_manager = None

    async def _load_events_from_redis(self):
        try:
            return await self.conn.lrange(self.name, 0, -1)
        except HTTPError:
            logger.exception("Peach collector raised an HTTP error")
            return []

    async def _send_to_peach(self, events) -> int:
        """
        Send events in batches.  As soon as one calls fails, abort
        :returns An int with how many we've many we've been able to send
        """
        num_sent = 0
        try:
            # fmt: off
            batches = [
                events[i: i + self.batch_size] for i in range(0, len(events), self.batch_size)
            ]
            # fmt: on
            for batch in batches:
                for event in batch:
                    self.collector.add_event(orjson.loads(event))
                await self.collector.send_events()
                num_sent += len(batch)
        except HTTPError:
            logger.exception("Peach collector raised an HTTP error")
        return num_sent

    async def _delete_from_redis(self, num_sent: int):
        """
        If this fails (after retries) we raise an error which causes drain to crash leaving lock
        in place
        """
        try:

            async def wrap():
                return await self.conn.ltrim(self.name, num_sent, -1)

            await retry_with_backoff(
                wrap,
                self.delete_from_redis_retries,
                self.delete_from_redis_retries_exponent,
                RedisError,
            )

        except RetryException:
            logger.exception(
                f"Failed deleting sent events. You need to manually issue LTRIM 0 {num_sent} before"
                f"deleting {self.lock_name}"
            )

    def _log_results(self, events: list, num_sent: int):
        if len(events):
            if num_sent == len(events):
                logger.info(f"{num_sent} events sent to Peach")
            else:
                if num_sent:
                    logger.warning(
                        f"We were only able to send {num_sent} out of {len(events)} events to Peach"
                    )
                else:
                    logger.warning(
                        f"We weren't able to send any events to Peach. {len(events)} waiting"
                    )
        else:
            logger.info("No events waiting to be sent to Peach")



from .__version__ import __description__, __title__, __version__
from ._collector import PeachCollector
from ._queue import PeachCollectorQueue

__all__ = [
    "__description__",
    "__title__",
    "__version__",
    "PeachCollector",
    "PeachCollectorQueue",
]



import asyncio
import logging
import time
from typing import Callable, Awaitable

logger = logging.getLogger(__name__)


def timestamp() -> int:
    """
    In milliseconds (UTC based)
    """
    return int(round(time.time() * 1000))


class RetryException(Exception):
    pass


async def retry_with_backoff(
    wrapped: Callable[[], Awaitable],
    retries: int = 5,
    retry_exponent: float = 3.0,
    exc_type: Exception = Exception,
):
    for trial in range(1, retries + 1):
        try:
            await wrapped()
            return
        except exc_type as e:
            sleep = retry_secs(retries, retry_exponent)[trial - 1]
            logger.warning(f"{wrapped} failed. Retrying in {sleep} secs", exc_info=e)
            await asyncio.sleep(sleep)
    raise RetryException(f"Failed to execute {wrapped}. Giving up after {trial} trials")


def retry_secs(retries, retry_exponent):
    return [int(i ** retry_exponent) for i in range(1, retries + 1)]




import pathlib
import setuptools
from peach_collector import __version__, __title__, __description__

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setuptools.setup(
    name=__title__,
    version=__version__,
    url="https://git.ebu.io/now/peach-collector-py",
    author="Matti Kotsalainen",
    author_email="kotsalainen@ebu.ch",
    description=__description__,
    long_description=README,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=(
        "aioredis",
        "aioredlock",
        "httpx",
        "orjson",
    ),
)
